import { functions, clone, forOwn, isFunction } from 'micro-dash';
import { HttpClientTestingModule, HttpTestingController } from '@angular/common/http/testing';
import { ApplicationRef, InjectFlags } from '@angular/core';
import { TestBed, fakeAsync, flushMicrotasks, tick, discardPeriodicTasks } from '@angular/core/testing';
import { assert, convertTime, wrapMethod } from 'js-core';
import { isArray } from 'rxjs/internal-compatibility';
import { NoopAnimationsModule } from '@angular/platform-browser/animations';
import { TestScheduler } from 'rxjs/testing';
import { __awaiter } from 'tslib';
import { CompileMetadataResolver } from '@angular/compiler';
import { BrowserDynamicTestingModule, platformBrowserDynamicTesting } from '@angular/platform-browser-dynamic/testing';

// adapted from https://github.com/ngneat/spectator/blob/e13c9554778bdb179dfc7235aedb4b3b90302850/projects/spectator/src/lib/mock.ts
/**
 * Creates a new object with jasmine spies for each method in `type`.
 *
 * ```ts
 * class Greeter {
 *   greet(name: string) {
 *     return `Hello, ${name}!`;
 *   }
 * }
 *
 * const spyObject = createSpyObject(Greeter);
 * spyObject.greet.and.returnValue("Hello, stub!");
 * expect(spyObject.greet("Eric")).toBe("Hello, stub!");
 * expectSingleCallAndReset(spyObject.greet, "Eric");
 * ```
 */
function createSpyObject(type) {
    const mock = {};
    for (let proto = type.prototype; proto; proto = Object.getPrototypeOf(proto)) {
        for (const key of functions(proto)) {
            mock[key] = jasmine.createSpy(key);
        }
    }
    return mock;
}

/**
 * Expects the spy to have been called the exact number of times as the argument lists provided, with those arguments, then resets the spy.
 *
 * ```ts
 * const spy = jasmine.createSpy();
 *
 * spy(1)
 * spy(2, 3)
 * spy()
 * expectCallsAndReset(spy, [1], [2, 3], []);
 * ```
 */
function expectCallsAndReset(spy, ...allArgs) {
    expect(spy).toHaveBeenCalledTimes(allArgs.length);
    expect(spy.calls.allArgs()).toEqual(allArgs);
    spy.calls.reset();
}

/**
 * Expects exactly one call to have been made to a jasmine spy, for it to have received the given arguments, then resets the spy.
 *
 * ```ts
 * const spy = jasmine.createSpy();
 *
 * spy(1, 2);
 * expectSingleCallAndReset(spy, 1, 2); // pass
 * expectSingleCallAndReset(spy, 1, 2); // fail
 *
 * spy(3);
 * spy(4);
 * expectSingleCallAndReset(spy, 3); // fail
 * ```
 */
function expectSingleCallAndReset(spy, ...params) {
    expect(spy).toHaveBeenCalledTimes(1);
    expect(spy).toHaveBeenCalledWith(...params);
    spy.calls.reset();
}

/** @hidden */
function extendMetadata(metadata, toAdd) {
    const result = clone(metadata);
    forOwn(toAdd, (val, key) => {
        result[key] = isArray(result[key]) ? result[key].concat(val) : val;
    });
    return result;
}
/**
 * Provides the foundation for an opinionated testing pattern.
 * - All test are run in the `fakeAsync` zone. This gives you full control over
 *   the timing of everything by default.
 * - Variables that are initialized for each test exist in a context that is
 *   thrown away, so they cannot leak between tests.
 * - Clearly separates initialization code from the test itself.
 * - Gives control over the simulated date & time with a single line of code.
 * - Always verifies no un-expected http requests were made during a test.
 * - Always discards periodic tasks at the end of each test to automatically
 *   avoid an error from the `fakeAsync` zone.
 *
 * This example tests a simple service that uses HttpClient, and is tested by
 * using `AngularContext` directly. More often `AngularContext` will be used a
 * super class. See {@link ComponentContext} for more common use cases.
 * ```ts
 *  // This is the class we will test.
 *  @Injectable({ providedIn: 'root' })
 *  class MemoriesService {
 *   constructor(private httpClient: HttpClient) {}
 *
 *   getLastYearToday(): Observable<any> {
 *     const datetime = new Date();
 *     datetime.setFullYear(datetime.getFullYear() - 1);
 *     const date = datetime.toISOString().split('T')[0];
 *     return this.httpClient.get(`http://example.com/post-from/${date}`);
 *   }
 * }
 *
 *  describe('MemoriesService', () => {
 *
 *   // Tests should have exactly 1 variable outside an "it": `ctx`.
 *   let ctx: AngularContext;
 *   beforeEach(() => {
 *     ctx = new AngularContext();
 *   });
 *
 *   it('requests a post from 1 year ago', () => {
 *
 *     // Before calling `run`, set up any context variables this test needs.
 *     ctx.startTime = new Date('2004-02-16T10:15:00.000Z');
 *
 *     // Pass the test itself as a callback to `run()`.
 *     ctx.run(() => {
 *       const httpBackend = ctx.inject(HttpTestingController);
 *       const myService = ctx.inject(MemoriesService);
 *
 *       myService.getLastYearToday().subscribe();
 *
 *       httpBackend.expectOne('http://example.com/post-from/2003-02-16');
 *     });
 *   });
 * });
 * ```
 */
class AngularContext {
    /**
     * @param moduleMetadata passed along to [TestBed.configureTestingModule()]{@linkcode https://angular.io/api/core/testing/TestBed#configureTestingModule}. Automatically includes {@link HttpClientTestingModule} for you.
     */
    constructor(moduleMetadata = {}) {
        /**
         * Set this before calling `run()` to mock the time at which the test starts.
         */
        this.startTime = new Date();
        TestBed.configureTestingModule(extendMetadata(moduleMetadata, { imports: [HttpClientTestingModule] }));
    }
    run(optionsOrTest, test) {
        let options = {};
        if (isFunction(optionsOrTest)) {
            test = optionsOrTest;
        }
        else {
            options = optionsOrTest;
        }
        jasmine.clock().install();
        fakeAsync(() => {
            jasmine.clock().mockDate(this.startTime);
            assert(test);
            this.init(options);
            try {
                test();
                this.verifyPostTestConditions();
            }
            finally {
                this.cleanUp();
            }
        })();
        jasmine.clock().uninstall();
    }
    /**
     * Gets a service or other injectable from the root injector. This implementation is a simple pass-through to [TestBed.inject()]{@linkcode https://angular.io/api/core/testing/TestBed#inject}, but subclasses may provide their own implementation. It is recommended to use this in your tests instead of using `TestBed` directly.
     */
    inject(token) {
        return TestBed.inject(token);
    }
    /**
     * Advance time and trigger change detection. It is common to call this with no arguments to trigger change detection without advancing time.
     *
     * @param unit The unit of time `amount` represents. Accepts anything described in `@s-libs/s-core`'s [TimeUnit]{@linkcode https://simontonsoftware.github.io/s-js-utils/typedoc/enums/timeunit.html} enum.
     */
    tick(amount = 0, unit = 'ms') {
        // To simulate real life, trigger change detection before processing macro tasks. To further simulate real life, wait until the micro task queue is empty.
        flushMicrotasks();
        this.runChangeDetection();
        tick(convertTime(amount, unit, 'ms'));
    }
    /**
     * This is a hook for subclasses to override. It is called during `run()`, before the `test()` callback. This implementation does nothing, but if you override this it is still recommended to call `super.init(options)` in case this implementation does something in the future.
     */
    init(_options) { }
    /** @hidden */
    runChangeDetection() {
        this.inject(ApplicationRef).tick();
    }
    /**
     * Runs post-test verifications. This base implementation runs [HttpTestingController#verify]{@linkcode https://angular.io/api/common/http/testing/HttpTestingController#verify}. Unlike {@link #cleanUp}, it is OK for this method to throw an error to indicate a violation.
     */
    verifyPostTestConditions() {
        this.inject(HttpTestingController).verify();
    }
    /**
     * Performs any cleanup needed at the end of each test. This base implementation calls [discardPeriodicTasks]{@linkcode https://angular.io/api/core/testing/discardPeriodicTasks} to avoid an error from the `fakeAsync` zone.
     */
    cleanUp() {
        discardPeriodicTasks();
    }
}

/** @hidden */
let initialStyles;
/**
 * When called at the right times, this speeds up your test suite by removing `<script>` tags that Angular component tests leave in the DOM. By default Angular dynamically adds `<script>` tags into the `<head>` of the document corresponding to component styles. In Jasmine tests they are never cleaned up. This can cause hundreds or even thousands of style tags to accumulate, making component tests slow down as your test suite runs. In Chrome debugger's performance tab, this time will show up under the activity "Recalculate Style".
 *
 * The first time this function is called it memorizes the `<script>` tags in the DOM, then each time it is called after it removes any that appeared since. A simple way to use this effectively is to call it in a `beforeEach()`.
 *
 * ```ts
 * beforeEach(() => {
 *   trimLeftoverStyles();
 * });
 * ```
 */
function trimLeftoverStyles() {
    const styles = new Set(Array.from(document.querySelectorAll('style')));
    if (initialStyles) {
        for (const style of styles) {
            if (!initialStyles.has(style)) {
                style.remove();
            }
        }
    }
    else {
        initialStyles = styles;
    }
}
/** @hidden */
function resetTrimLeftoverStyles() {
    initialStyles = undefined;
}

/**
 * A superclass to set up testing contexts for components. This is a foundation for an opinionated testing pattern, including everything described in {@link AngularContext}. A very simple example:
 *
 * ```ts
 * @Component({ template: 'Hello, {{name}}!' })
 * class GreeterComponent {
 *   @Input() name!: string;
 * }
 *
 * class GreeterComponentContext extends ComponentContext {
 *   protected componentType = GreeterComponent;
 * }
 *
 * describe('ComponentContext', () => {
 *   let ctx: GreeterComponentContext;
 *   beforeEach(() => {
 *     ctx = new GreeterComponentContext();
 *   });
 *
 *   it('greets you by name', () => {
 *     ctx.run({ input: { name: 'World' } }, () => {
 *       expect(ctx.fixture.nativeElement.textContent).toBe('Hello, World!');
 *     });
 *   });
 * })
 * ```
 *
 * This class integrates {@link trimLeftoverStyles} to speed up your test suite, as described in the docs for that function.
 */
class ComponentContext extends AngularContext {
    /**
     * @param moduleMetadata passed along to [TestBed.configureTestingModule()]{@linkcode https://angular.io/api/core/testing/TestBed#configureTestingModule}. Automatically includes {@link NoopAnimationsModule} and {@link HttpClientTestingModule} for you.
     */
    constructor(moduleMetadata = {}) {
        super(extendMetadata(moduleMetadata, { imports: [NoopAnimationsModule] }));
    }
    /**
     * Constructs and initializes your component. Called during `.run()` before it executes the rest of your test. Runs in the same `fakeAsync` zone as the rest of your test.
     */
    init(options) {
        trimLeftoverStyles();
        super.init(options);
        this.fixture = TestBed.createComponent(this.componentType);
        Object.assign(this.fixture.componentInstance, options.input);
        this.fixture.detectChanges();
        this.tick();
    }
    /** @hidden */
    runChangeDetection() {
        this.fixture.detectChanges();
    }
    /**
     * Performs any cleanup needed at the end of each test. This implementation destroys {@link fixture} and calls the super implementation.
     */
    cleanUp() {
        this.fixture.destroy();
        super.cleanUp();
    }
}

/**
 * Designed for temporary use when a test using `fakeAsync` is giving you "Error: X timer(s) still in the queue", and you don't know what they are. Call this at the top of the test to see everything that adds a timer printed to the JS console. In Chrome you can click the function in the console output to be taken directly to its definition in the Sources tab.
 *
 * Be sure to call this _inside_ the test, not before, to get the most useful output.
 *
 * @returns a function to stop the logging
 *
 * ```ts
 * it('definitely does not set any timeouts', fakeAsync(() => {
 *   const stopLogging = logTimers();
 *
 *   doTheRestOfMyTest();
 *
 *   stopLogging();
 * }));
 * ```
 */
function logTimers() {
    const restoreTimeouts = wrapMethod(window, 'setTimeout', {
        before(...args) {
            console.log('setTimeout(', ...args, ')');
        },
    });
    const restoreIntervals = wrapMethod(window, 'setInterval', {
        before(...args) {
            console.log('setInterval(', ...args, ')');
        },
    });
    return () => {
        restoreTimeouts();
        restoreIntervals();
    };
}

/**
 * A helper to take some boilerplate out of RxJS's new marble testing API for jasmine.
 *
 * ```ts
 * it("doubles numbers", marbleTest(({ cold, expectObservable }) => {
 *   const source = cold("-1-2-3-|");
 *   const expected = "   -2-4-6-|");
 *   expectObservable(source.pipe(double())).toBe(expected);
 * }));
 * ```
 */
function marbleTest(callback) {
    const testScheduler = new TestScheduler((actual, expected) => {
        expect(actual).toEqual(expected);
    });
    return () => testScheduler.run((helpers) => callback(Object.assign(Object.assign({}, helpers), { testScheduler })));
}

/**
 * Use this to speed up your angular test suite. Normally when you call `TestBed.compileComponents()`, it recompiles all the configured components even if it has compiled them before in a previous test. AoT compilation would be a great fix, but is not well supported/documented for tests yet. This function fills the gap by precompiling the components as if by AoT and setting up `TestBed` to use them.
 *
 * ```ts
 * // let's assume `AppModule` declares or imports a `HelloWorldComponent`
 * precompileForTests([AppModule]);
 *
 * // Everything below here is the same as normal. Just add the line above.
 *
 * describe("AppComponent", () => {
 *   it("says hello", async () => {
 *     TestBed.configureTestingModule({ declarations: [HelloWorldComponent] });
 *     await TestBed.compileComponents(); // <- this line is faster
 *     const fixture = TestBed.createComponent(HelloWorldComponent);
 *     expect(fixture.nativeElement.textContent).toContain("Hello, world!");
 *   });
 * });
 * ```
 *
 * Note that this uses `TestBed.initTestEnvironment` to set up the precompiled components, so you will not be able to use it yourself at the same time.
 *
 * @deprecated This is no longer needed with Ivy.
 */
function precompileForTests(modules, skipModules = []) {
    beforeAll(() => __awaiter(this, void 0, void 0, function* () {
        // technique modeled from https://github.com/angular/angular/blob/11325bad4ab786a07e52ff380c00622fda11c0b7/packages/core/test/linker/jit_summaries_integration_spec.ts#L83
        const metadataResolver = optionallyInject(CompileMetadataResolver);
        if (!metadataResolver) {
            console.warn(`precompileForTests() is no longer needed with Ivy. Skipping.`);
            return;
        }
        TestBed.configureTestingModule({ imports: modules });
        yield TestBed.compileComponents();
        TestBed.resetTestEnvironment();
        TestBed.initTestEnvironment(BrowserDynamicTestingModule, platformBrowserDynamicTesting(), () => extractAotSummaries(metadataResolver, modules, new Set(skipModules)));
    }));
}
// TODO: this could be a public function
/** @hidden */
function optionallyInject(token) {
    return TestBed.inject(token, null, InjectFlags.Optional);
}
/** @hidden */
function extractAotSummaries(metadataResolver, modules, skipModules) {
    return modules.map((module) => () => extractAotSummaries0(metadataResolver, module, skipModules));
}
/** @hidden */
function extractAotSummaries0(metadataResolver, module, skipModules) {
    if (skipModules.has(module)) {
        return [];
    }
    skipModules.add(module);
    const moduleMetadata = metadataResolver.getNgModuleMetadata(module);
    return [
        metadataResolver.getNgModuleSummary(module),
        () => extractDirectives(moduleMetadata, metadataResolver),
        () => extractPipes(moduleMetadata, metadataResolver),
        () => extractAotSummaries(metadataResolver, extractImports(moduleMetadata), skipModules),
    ];
}
/** @hidden */
function extractDirectives(moduleMetadata, metadataResolver) {
    return moduleMetadata.declaredDirectives.map((directive) => metadataResolver.getDirectiveSummary(directive.reference));
}
/** @hidden */
function extractPipes(moduleMetadata, metadataResolver) {
    return moduleMetadata.declaredPipes.map((pipe) => metadataResolver.getPipeSummary(pipe.reference));
}
/** @hidden */
function extractImports(moduleMetadata) {
    return moduleMetadata.importedModules.map((importedModule) => importedModule.type.reference);
}

/*
 * Public API Surface of ng-dev
 */

/**
 * Generated bundle index. Do not edit.
 */

export { AngularContext, ComponentContext, createSpyObject, expectCallsAndReset, expectSingleCallAndReset, logTimers, marbleTest, precompileForTests, trimLeftoverStyles };
//# sourceMappingURL=s-libs-ng-dev.js.map
